from strictyaml import load
from commandlib import Command, python_bin
import hitchpylibrarytoolkit


class ProjectDocumentation:
    def __init__(
        self,
        storybook,
        project_path,
        publish_path,
        project_name,
        github_address,
        image="",
    ):
        self._storybook = storybook
        self._project_path = project_path
        self._publish_path = publish_path
        self._project_name = project_name
        self._github_address = github_address
        self._project_slug = project_name.lower()
        self._image = image

    def _readme_intro(self):
        return (
            "# {project_name}\n"
            "\n"
            "[![Main branch status](https://github.com/{github_address}/actions/workflows/regression.yml/badge.svg)](https://github.com/{github_address}/actions/workflows/regression.yml)"
        ).format(
            project_name=self._project_name,
            github_address=self._github_address,
        )

    def _docs_intro(self):
        return (
            "---\n"
            "title: {project_name}\n"
            "---\n"
            "\n{image}\n\n"
            '<img alt="GitHub Repo stars" src="https://img.shields.io/github/stars/{github_address}?style=social">'
            '<img alt="PyPI - Downloads" src="https://img.shields.io/pypi/dm/{project_slug}">'
        ).format(
            image=self._image,
            project_name=self._project_name,
            github_address=self._github_address,
            project_slug=self._project_slug,
        )

    def _title(self, filepath):
        try:
            assert len(filepath.text().split("---")) >= 3, "{} doesn't have ---".format(
                filepath
            )
            return load(filepath.text().split("---")[1]).data.get("title", "misc")
        except UnicodeDecodeError:
            return None
        except AssertionError:
            return None

    def _contents(self, main_folder, folder, readme):
        markdown = ""
        for filepath in sorted(main_folder.joinpath(folder).listdir()):
            if filepath.name != "index.md":
                title = self._title(filepath)

                if title is not None:
                    path = filepath.relpath(main_folder).stripext()

                    markdown += "- [{}]({})\n".format(
                        title,
                        "https://hitchdev.com/{}/".format(self._project_slug) + path
                        if readme
                        else path,
                    )
        return markdown

    def generate(self, readme=False):
        dirtempl = python_bin.dirtempl.in_dir(self._project_path / "docs")
        doc_src = self._project_path / "docs" / "src"
        snippets_path = self._project_path / "docs" / "snippets"
        git = Command("git").in_dir(self._project_path)

        dest_path = self._publish_path

        if snippets_path.exists():
            snippets_path.rmtree()
        snippets_path.mkdir()

        if dest_path.exists():
            dest_path.rmtree()

        snippets_path.joinpath("quickstart.txt").write_text(
            self._storybook.with_documentation(
                self._project_path.joinpath("hitch", "docstory.yml").text(),
                extra={"in_interpreter": True, "include_title": False},
            )
            .named("Quickstart")
            .documentation()
        )

        self._generate_storydocs(
            self._project_path.joinpath("hitch", "docstory.yml").text(),
            doc_src.joinpath("using"),
            self._storybook,
        )

        snippets_path.joinpath("intro.txt").write_text(
            self._readme_intro() if readme else self._docs_intro()
        )

        for folder in ["why", "approach", "why-not", "using"]:
            snippets_path.joinpath(f"{folder}-contents.txt").write_text(
                self._contents(doc_src, folder, readme=readme)
            )

        for folder in ["why", "approach", "why-not", "using"]:
            snippets_path.joinpath(f"{folder}-index-contents.txt").write_text(
                self._contents(doc_src / folder, "", readme=readme)
            )

        dirtempl("--snippets", snippets_path, doc_src, dest_path).run()

        dest_path.joinpath("changelog.md").write_text(
            hitchpylibrarytoolkit.docgen.changelog(self._project_path)
        )

    def _generate_storydocs(self, docstory, docpath, storybook):
        storydocs = storybook.with_documentation(
            docstory,
            extra={"in_interpreter": False, "include_title": True},
        )

        for story in storydocs.ordered_by_file():
            docfilename = story.info.get("docs")
            if docfilename is not None:
                docpath.joinpath(docfilename + ".md").write_text(story.documentation())



from commandlib import Command, python
from path import Path
import os


def deploy(project_name, github_path, temp_path, testpypi=False, dryrun=False, python_cmd=None):
    git = Command("git")
    
    if python_cmd is None:
        python_cmd = python

    if temp_path.joinpath(project_name).exists():
        temp_path.joinpath(project_name).rmtree()

    Path("/root/.ssh/known_hosts").write_text(
        Command("ssh-keyscan", "github.com").output()
    )

    git("clone", "git@github.com:{}.git".format(github_path)).in_dir(temp_path).run()
    project = temp_path / project_name
    version = project.joinpath("VERSION").text().rstrip()
    initpy = project.joinpath(project_name, "__init__.py")
    original_initpy_contents = initpy.bytes().decode("utf8")
    initpy.write_text(original_initpy_contents.replace("DEVELOPMENT_VERSION", version))
    python_cmd("-m", "pip", "wheel", ".", "-w", "dist").in_dir(project).run()
    python_cmd("-m", "build", "--sdist").in_dir(project).run()
    initpy.write_text(original_initpy_contents)

    # Upload to pypi
    wheel_args = ["-m", "twine", "upload"]
    if testpypi:
        wheel_args += ["--repository", "testpypi"]
    wheel_args += ["dist/{}-{}-py3-none-any.whl".format(project_name, version)]

    if not dryrun:
        python_cmd(*wheel_args).in_dir(project).with_env(
            TWINE_USERNAME="__token__",
            TWINE_PASSWORD=os.getenv("PYPITOKEN"),
        ).run()

    sdist_args = ["-m", "twine", "upload"]
    if testpypi:
        sdist_args += ["--repository", "testpypi"]
    sdist_args += ["dist/{0}-{1}.tar.gz".format(project_name, version)]
    
    if not dryrun:
        python_cmd(*sdist_args).in_dir(project).with_env(
            TWINE_USERNAME="__token__",
            TWINE_PASSWORD=os.getenv("PYPITOKEN"),
        ).run()

    # Clean up
    temp_path.joinpath(project_name).rmtree()



from hitchpylibrarytoolkit.toolkit import ProjectToolkitV2  # noqa: F401
from hitchpylibrarytoolkit.toolkit import ProjectToolkit  # noqa: F401
from hitchpylibrarytoolkit.exceptions import ToolkitError  # noqa: F401


__version__ = "0.6.14"



from hitchpylibrarytoolkit.deploy import deploy
from hitchpylibrarytoolkit.docgen import docgen
from hitchpylibrarytoolkit.docgen import readmegen
from hitchpylibrarytoolkit.exceptions import ToolkitError
from hitchpylibrarytoolkit.project_docs import ProjectDocumentation
from hitchpylibrarytoolkit import pyenv
import hitchpylibrarytoolkit
from commandlib import python, python_bin, Command, CommandError
from hitchstory import StoryCollection
from pathquery import pathquery
from path import Path
import os


class Directories:
    gen = Path("/gen")
    key = Path("/src/hitch/")
    project = Path("/src/")
    share = Path("/gen")


class ProjectToolkit(object):
    def __init__(self, project_name, paths):
        self._project_name = project_name
        self._path = paths

    @property
    def current_version(self):
        return self._path.project.joinpath("VERSION").text().rstrip()

    def bdd(self, engine, keywords):
        """Run individual story matching key words."""
        self._stories(engine).only_uninherited().shortcut(*keywords).play()

    def regression(self, engine):
        self._stories(engine).only_uninherited().ordered_by_name().play()

    def _stories(self, engine):
        return StoryCollection(
            pathquery(self._path.key / "story").ext("story"),
            engine,
        )

    def prepdeploy(self, engine, version):
        """Prepare a deployment with new version, README and docs, but do not push."""
        git = Command("git").in_dir(self._path.project)
        version_file = self._path.project.joinpath("VERSION")
        old_version = version_file.text().rstrip()
        from packaging.version import parse

        if not parse(version) > parse():
            raise ToolkitError(
                "New version {} must be above old version {}".format(
                    version, old_version
                )
            )
        version_file.write_text(version)
        self.readmegen(engine)
        self.docgen(engine)
        self.validate_readmegen(engine)
        self.validate_docgen(engine)
        git("add", ".").run()
        git(
            "commit", "-m", "RELEASE: Version {0} -> {1}".format(old_version, version)
        ).run()
        git("tag", "-a", version, "-m", "Version {0}".format(version)).run()

    def pushdeploy(self):
        """Push a new deployment to pypi."""
        git = Command("git").in_dir(self._path.project)
        version = self._path.project.joinpath("VERSION").text().rstrip()
        git("push", "--follow-tags").run()

        # Set __version__ variable in __init__.py, build sdist and put it back
        initpy = self._path.project.joinpath(self._project_name, "__init__.py")
        original_initpy_contents = initpy.bytes().decode("utf8")
        initpy.write_text(
            original_initpy_contents.replace("DEVELOPMENT_VERSION", version)
        )
        python("setup.py", "sdist").in_dir(self._path.project).run()
        initpy.write_text(original_initpy_contents)

        # Upload to pypi
        python(
            "-m",
            "twine",
            "upload",
            "dist/{0}-{1}.tar.gz".format(self._project_name, version),
        ).in_dir(self._path.project).run()
        

    def lint(self, exclude=None):
        try:
            if exclude is None:
                exclude = "__init__.py"
            python("-m", "flake8")(
                self._path.project.joinpath(self._project_name),
                "--max-line-length=100",
                "--exclude={}".format(",".join(exclude)),
                "--ignore=E203,W503",  # Ignore list[expression1 : expression2] failures
            ).run()
            python("-m", "flake8")(
                self._path.project.joinpath("hitch", "key.py"),
                "--max-line-length=100",
                "--ignore=E203,W503",
            ).run()
        except CommandError:
            raise ToolkitError("Linter failed")

    def reformat(self):
        python_bin.black(self._path.project / self._project_name).run()
        python_bin.black(self._path.project / "hitch" / "key.py").run()

    def validate_reformatting(self):
        try:
            python_bin.black(self._path.project / self._project_name, "--check").run()
            python_bin.black(self._path.project / "hitch" / "key.py", "--check").run()
        except CommandError:
            raise ToolkitError("Black failed")

    def docgen(self, engine):
        docgen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.project / "docs" / "public",
            self._path.gen / "tmpdocs",
        )

    def validate_docgen(self, engine):
        docgen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.project / "docs" / "public",
            self._path.gen / "tmpdocs",
            check=True,
        )

    def readmegen(self, engine):
        readmegen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.gen,
            self._project_name,
        )

    def validate_readmegen(self, engine):
        readmegen(
            self._stories(engine),
            self._path.project,
            self._path.key / "story",
            self._path.gen,
            self._project_name,
            check=True,
        )


class ProjectToolkitV2(ProjectToolkit):
    def __init__(self, name, slug, github_address, image=""):
        self._name = name
        self._project_name = slug
        self._github_address = github_address
        self.DIR = Directories()
        self._path = self.DIR
        self._image = image

    def devenv(self):
        env = pyenv.DevelopmentVirtualenv(
            pyenv.Pyenv(self.DIR.gen / "pyenv"),
            self.DIR.project.joinpath("hitch", "devenv.yml"),
            self.DIR.project.joinpath("hitch", "debugrequirements.txt"),
            self.DIR.project,
            self.DIR.project.joinpath("pyproject.toml").text(),
        )
        env.ensure_built()
        return env

    def deploy(self, testpypi=False):
        relenv = pyenv.ReleaseVirtualenv(
            pyenv.Pyenv(self.DIR.gen / "pyenv")
        )
        relenv.ensure_built()
        
        hitchpylibrarytoolkit.deploy.deploy(
            self._project_name,
            self._github_address,
            self.DIR.gen,
            testpypi=testpypi,
            dryrun=False,
            python_cmd = Command(relenv.python_path),
        )
    
    def package_test(self):
        relenv = pyenv.ReleaseVirtualenv(
            pyenv.Pyenv(self.DIR.gen / "pyenv")
        )
        relenv.ensure_built()

        hitchpylibrarytoolkit.deploy.deploy(
            self._project_name,
            self._github_address,
            self.DIR.gen,
            testpypi=True,
            dryrun=True,
            python_cmd = Command(relenv.python_path),
        )

    def draft_docs(self, storybook):
        ProjectDocumentation(
            storybook,
            self.DIR.project,
            self.DIR.project / "docs" / "draft",
            self._name,
            self._github_address,
            image=self._image,
        ).generate()

    def publish(self, storybook):
        if self.DIR.gen.joinpath(self._project_name).exists():
            self.DIR.gen.joinpath(self._project_name).rmtree()

        Path("/root/.ssh/known_hosts").write_text(
            Command("ssh-keyscan", "github.com").output()
        )
        
        print(os.getenv("CI"))
        if os.getenv("CI").lower() == "true":
            Command(
                "git", "clone", "https://{}@github.com/{}.git".format(
                    os.getenv("GITHUBTOKEN").rstrip(),
                    self._github_address,
                )
            ).in_dir(self.DIR.gen).run()
        else:
            Command(
                "git", "clone", "git@github.com:{}.git".format(self._github_address)
            ).in_dir(self.DIR.gen).run()

        git = Command("git").in_dir(self.DIR.gen / self._project_name)
        git("config", "user.name", "Bot").run()
        git("config", "user.email", "bot@hitchdev.com").run()
        git("rm", "-r", "docs/public").run()

        ProjectDocumentation(
            storybook,
            self.DIR.project,
            self.DIR.gen / self._project_name / "docs" / "public",
            self._name,
            self._github_address,
            image=self._image,
        ).generate()

        ProjectDocumentation(
            storybook,
            self.DIR.project,
            self.DIR.gen / self._project_name / "docs" / "draft",
            self._name,
            self._github_address,
            image="",
        ).generate(readme=True)

        self.DIR.gen.joinpath(self._project_name, "docs", "draft", "index.md").copy(
            self.DIR.gen / self._project_name / "README.md"
        )
        self.DIR.gen.joinpath(self._project_name, "docs", "draft", "changelog.md").copy(
            self.DIR.gen / self._project_name / "CHANGELOG.md"
        )

        git("add", self.DIR.gen / self._project_name).run()
        git("commit", "-m", "DOCS : Regenerated docs.").in_dir(
            self.DIR.gen / self._project_name
        ).run()

        git("push").run()
